"""
PrimalScheme: a primer3 wrapper for designing multiplex primer schemes

Copyright (C) 2020 Joshua Quick and Andrew Smith
www.github.com/aresti/primalscheme

This module contains MultiplexScheme.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>
"""

import logging

from primalscheme import config
from primalscheme.align import FailedAlignmentError
from primalscheme.region import Region, NoSuitablePrimersError

logger = logging.getLogger("primalscheme")


class MultiplexScheme:
    """A complete multiplex primer scheme."""

    def __init__(
        self,
        references,
        prefix=config.PREFIX,
        amplicon_size_min=config.AMPLICON_SIZE_MIN,
        amplicon_size_max=config.AMPLICON_SIZE_MAX,
        target_overlap=config.TARGET_OVERLAP,
        primary_only=False,
        high_gc=False,
        progress_tracker=None,
    ):
        """Init MultiplexScheme."""

        if amplicon_size_min > amplicon_size_max:
            raise ValueError("amplicon_size_min cannot exceed amplicon_size_max")

        self.prefix = prefix
        self.amplicon_size_min = amplicon_size_min
        self.amplicon_size_max = amplicon_size_max
        self.target_overlap = target_overlap
        self.primary_only = primary_only
        self.high_gc = high_gc
        self.progress_tracker = progress_tracker

        self.regions = []
        self.references = references
        self.primary_ref = references[0]
        self.secondary_refs = references[1:]
        self.excluded_refs = []
        self.ref_len = len(self.primary_ref)
        self._max_failed_aln = int(
            config.MAX_ALN_GAP_PERCENT * self.ref_len / config.STEP_DISTANCE
        )
        self.considered = 0

        # config switching
        config_key = "HIGH_GC" if high_gc else "DEFAULT"
        config.PRIMER_SIZE_RANGE = config.PRIMER_SIZE_RANGES[config_key]
        config.PRIMER_GC_RANGE = config.PRIMER_GC_RANGES[config_key]

        # progress & logging setup
        if self.progress_tracker:
            self.progress_tracker.end = self.ref_len

        logger.debug(str(self))

    def __str__(self):
        lines = [
            "MuliplexScheme",
            f"Prefix: {self.prefix}",
            f"Amplicon size range: {self.amplicon_size_min} - {self.amplicon_size_max}",
            f"Target overlap: {self.target_overlap}",
            f"Primary only (pinned): {self.primary_only}",
            "References:",
        ]
        lines.extend(f" - {ref.id}" for ref in self.references)
        return "\n".join(lines)

    @property
    def region_count(self):
        """The number of regions in the scheme."""
        return len(self.regions)

    @property
    def _region_num(self):
        """The next/pending region number."""
        return len(self.regions) + 1

    @property
    def _prev(self):
        """The previous region."""
        return self.regions[-1] if self.regions else None

    @property
    def _prev_in_pool(self):
        """The previous region in the same pool as the next."""
        return self.regions[-2] if self.region_count > 1 else None

    @property
    def _remaining_distance(self):
        """The distance between the last region and the primary reference end."""
        if self.regions:
            return self.ref_len - self._prev.right.start
        return self.ref_len

    @property
    def _is_last_region(self):
        """Is there space for only one more region?"""
        return self._remaining_distance <= self.amplicon_size_max

    @property
    def _left_limit(self):
        """The left limit for the next region."""
        if self._region_num == 1:
            return 0
        elif self._region_num == 2 or (
            self._prev.left.start > self._prev_in_pool.right.start
        ):
            # first region in second pool, or we have a gap. Force progress.
            return self._prev.left.end + 1
        else:
            # default case (just don't crash into prev in pool)
            return self._prev_in_pool.right.start + 1

    @property
    def _slice_start(self):
        """The ideal slice start position for the next region."""

        if self._region_num == 1:
            return 0

        insert_start = self._prev.right.end - self.target_overlap - 1
        desired_slice_start = (
            insert_start
            - config.PRIMER_SIZE_RANGE.max
            - (self.amplicon_size_max - self.amplicon_size_min)
        )

        # if target overlap is impossible, take left_limit
        slice_start = max(desired_slice_start, self._left_limit)

        if self._is_last_region:
            # if forced to choose, take a gap at the end
            right_aligned = self.ref_len - self.amplicon_size_max
            return min(slice_start, right_aligned)

        return slice_start

    @property
    def primers(self):
        """Return a list of all primers in the scheme."""
        primers = []
        for region in [region for region in self.regions]:
            primers.append(region.left)
            primers.append(region.right)
        return primers

    def primers_in_pool(self, pool):
        """Return a list of all primers in a pool."""
        return [primer for primer in self.primers if primer.pool == pool]

    def design_scheme(self):
        """Design a multiplex primer scheme."""
        is_last = False
        while not is_last:
            is_last = self._is_last_region
            region = Region(self._region_num, self, self._left_limit, self._slice_start)
            try:
                region.find_primers()
            except NoSuitablePrimersError as exc:
                if self._region_num == 1:
                    raise exc
                break
            except FailedAlignmentError as exc:
                self._exclude_reference(exc.reference)
                continue

            self.regions.append(region)

            if self.progress_tracker:
                self.progress_tracker.region_num = self._region_num
                self.progress_tracker.goto(region.right.start)

            if self._is_last_region and self._slice_start < self._left_limit:
                # additional region no longer possible due to small ref, relative to
                # amplicon size, or significant right stepping of prev region.
                break

        if self.progress_tracker:
            self.progress_tracker.goto(self.progress_tracker.end)
            self.progress_tracker.finish()

    def _exclude_reference(self, reference):
        """Exclude a secondary reference."""
        for i, ref in enumerate(self.secondary_refs):
            if ref.id == reference.id:
                self.excluded_refs.append(self.secondary_refs.pop(i))
        self.progress_tracker.interrupt()
        logger.info(f"Excluding reference {reference.id}: unable to align.")

    def add_considered(self, num):
        """Increment number of considered primers; update progress tracker."""
        self.considered += num
        if self.progress_tracker:
            self.progress_tracker.considered = self.considered



"""
PrimalScheme: a primer3 wrapper for designing multiplex primer schemes

Copyright (C) 2020 Joshua Quick and Andrew Smith
www.github.com/aresti/primalscheme

This module contains the Window and Region classes.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>
"""

import logging
import primer3

from operator import attrgetter
from Bio.SeqRecord import SeqRecord
from Bio.Align import MultipleSeqAlignment

from primalscheme import config
from primalscheme.align import align_secondary_reference, FailedAlignmentError
from primalscheme.primer import design_primers, primer_thermo_filter, Direction

logger = logging.getLogger("primalscheme")


class Window:
    """
    A sliding window representing slice coordinates
    against the primary reference.
    """

    def __init__(self, scheme, left_limit, slice_start, right_limit=None):
        """Init Window."""
        self.scheme = scheme
        self.left_limit = left_limit
        self.slice_start = slice_start
        self.right_limit = right_limit or len(scheme.primary_ref.seq)

        self._initial_slice_start = self.slice_start
        self.left_flank_msa = None
        self.right_flank_msa = None

        logger.debug(
            f"Window: left_limit {left_limit}, slice_start {slice_start}, "
            f"right_limit {self.right_limit}"
        )

        # check bounds
        if slice_start < left_limit or self.slice_end > self.right_limit:
            raise SliceOutOfBoundsError("The window slice is out of bounds.")

    def step_left(self):
        """Step the slice start position left, raise if limit reached"""
        distance = config.STEP_DISTANCE
        if (self.slice_start - distance) < self.left_limit:
            logger.debug("Left limit reached")
            raise SliceOutOfBoundsError("Left window limit reached.")
        self.slice_start -= distance
        self.left_flank_msa, self.right_flank_msa = None, None
        logger.debug(f"Stepping left to {self.slice_start}")

    def step_right(self):
        """Step the slice start position right, raise if limit reached"""
        distance = config.STEP_DISTANCE
        if (self.slice_end + distance) > self.right_limit:
            logger.debug("Right limit reached")
            raise SliceOutOfBoundsError("Right window limit reached.")
        self.slice_start += distance
        self.left_flank_msa, self.right_flank_msa = None, None
        logger.debug(f"Stepping right to {self.slice_start}")

    def reset_slice(self):
        """Reset the slice start position to its initial value"""
        self.slice_start = self._initial_slice_start

    @property
    def slice_end(self):
        """The slice end position"""
        return self.slice_start + self.scheme.amplicon_size_max - 1

    @property
    def ref_slice(self):
        """The reference sequence for the slice"""
        return self.scheme.primary_ref[self.slice_start : self.slice_end + 1]

    @property
    def flank_size(self):
        """The size of the slice flanks, where possible primers could be located"""
        return (
            int((self.scheme.amplicon_size_max - self.scheme.amplicon_size_min) / 2)
            + config.PRIMER_SIZE_RANGE.max
        )

    @property
    def left_flank(self):
        """The reference sequence for the left flank"""
        return self.ref_slice[: self.flank_size]

    @property
    def right_flank(self):
        """The reference sequence for the right flank"""
        return self.ref_slice[-self.flank_size :]

    def align_flanks(self):
        """Perform multiple seq alignment of all refs for flanks."""
        self.left_flank_msa = MultipleSeqAlignment([self.left_flank])
        right_rev = SeqRecord(
            self.right_flank.reverse_complement().seq, id=self.scheme.primary_ref.id
        )
        self.right_flank_msa = MultipleSeqAlignment([right_rev])
        for ref in self.scheme.secondary_refs:
            self.left_flank_msa.append(align_secondary_reference(self.left_flank, ref))
            ra = align_secondary_reference(self.right_flank, ref)
            self.right_flank_msa.append(
                SeqRecord(ra.reverse_complement().seq, id=ra.id)
            )


class Region(Window):
    """
    A region forming part of a tiling amplicon scheme.
    """

    def __init__(self, region_num, *args, **kwargs):
        """Init Region."""
        self.region_num = region_num
        self.pool = 1 if self.region_num % 2 == 1 else 2
        super().__init__(*args, **kwargs)

        self.left = None
        self.right = None
        self.left_candidates = []
        self.right_candidates = []
        self.exhausted_left_stepping = False
        self.failed_alignment_count = 0

        logger.debug(f"Region {region_num}, pool {self.pool}")

    @property
    def product_size(self):
        """The product size for the picked left/right primers"""
        if self.left and self.right:
            return self.right.start - self.left.start + 1
        return None

    def find_primers(self):
        """Find primers for this region"""
        while True:
            try:
                self._find_primers_for_slice()
                return
            except FailedAlignmentError as exc:
                self.failed_alignment_count += 1
                if self.failed_alignment_count > self.scheme._max_failed_aln:
                    raise exc
                self._try_stepping()
            except NoSuitablePrimersError:
                self._try_stepping()

    def _try_stepping(self):
        """Step region left, or right (from initial start) once left limit reached"""
        if self.exhausted_left_stepping:
            try:
                self.step_right()
            except SliceOutOfBoundsError:
                raise NoSuitablePrimersError("Right limit reached.")
        else:
            try:
                self.step_left()
            except SliceOutOfBoundsError:
                self.reset_slice()
                self.exhausted_left_stepping = True

    def _find_primers_for_slice(self):
        """Try to find suitable primers for the current slice"""
        logger.debug(
            f"Finding primers for slice {self.slice_start} to {self.slice_end}"
        )

        # Align flanks at this position
        self.align_flanks()

        # Design primers for the left and right flanks
        candidates = design_primers(
            self.left_flank_msa,
            Direction.LEFT,
            self.pool,
            offset=self.slice_start,
            primary_only=self.scheme.primary_only,
        )
        candidates.extend(
            design_primers(
                self.right_flank_msa,
                Direction.RIGHT,
                self.pool,
                offset=self.slice_end - self.flank_size + 1,
                primary_only=self.scheme.primary_only,
            )
        )

        # Update considered (trigger progress update)
        self.scheme.add_considered(len(candidates))

        # Hard thermo filter
        filtered_candidates = [
            primer for primer in candidates if primer_thermo_filter(primer)
        ]

        # Check mismatch threshold
        passing_candidates = []
        for primer in filtered_candidates:
            if max(primer.mismatch_counts) <= config.PRIMER_MAX_MISMATCHES:
                passing_candidates.append(primer)
        candidates = passing_candidates

        # Pull out left and right from passing candidates
        self.left_candidates = [
            primer for primer in candidates if primer.direction == Direction.LEFT
        ]
        self.right_candidates = [
            primer for primer in candidates if primer.direction == Direction.RIGHT
        ]

        if not (self.left_candidates and self.right_candidates):
            raise NoSuitablePrimersError(
                "Unable to find at least one suitable pair of candidate primers."
            )

        # Pick best-scoring left and right candidates
        self._pick_pair()

    def _sort_candidates(self, candidates):
        """
        Sort candidates by penalty, start (higher), seq.
        seq is necessary to maintain deterministic output.
        """
        candidates.sort(key=attrgetter("seq"))
        candidates.sort(key=attrgetter("start"), reverse=True)
        candidates.sort(key=attrgetter("combined_penalty"))

    def _pick_pair(self):
        """Pick the best scoring left and right primer for the region"""
        self._sort_candidates(self.left_candidates)
        self._sort_candidates(self.right_candidates)
        self.left = self._pick_candidate(self.left_candidates)
        self.right = self._pick_candidate(self.right_candidates)

        if logger.level >= logging.DEBUG:
            self._log_debug(Direction.LEFT)
            self._log_debug(Direction.RIGHT)

    def _pick_candidate(self, candidates):
        """Pick the best scoring candidate that passes a same-pool heterodimer check"""
        for candidate in candidates:
            if not self._check_for_heterodimers(candidate):
                return candidate
        raise NoSuitablePrimersError(
            "All candidates form stable heterodimers with existing primers "
            "in this pool."
        )

    def _check_for_heterodimers(self, candidate):
        """
        Return True if candidate primer forms stable heterodimer with
        an existing primer in the same pool.
        """
        for existing in self.scheme.primers_in_pool(self.pool):
            thermo_end = primer3.bindings.calcEndStability(
                candidate.seq,
                existing.seq,
                mv_conc=config.MV_CONC,
                dv_conc=config.DV_CONC,
                dna_conc=config.DNA_CONC,
                dntp_conc=config.DNTP_CONC,
                temp_c=config.TEMP_C,
            )
            if thermo_end.dg / 1000 < config.HETERODIMER_DG_THRESHOLD:
                logger.debug(
                    f"Primer interaction between {candidate.seq} and {existing.seq} "
                    f"predicted with a ∆G of {thermo_end.dg / 1000:.2f} kcal/mol"
                )
                candidate.interacts_with = existing
                return True
        return False

    def _log_debug(self, direction):
        """Log detailed debug info for the region"""
        if direction == Direction.LEFT:
            logger.debug(f"Left region flank MSA: {self.left_flank_msa}")
            candidates = self.left_candidates
            picked = self.left
        else:
            logger.debug(f"Right region flank MSA: {self.right_flank_msa}")
            candidates = self.right_candidates
            picked = self.right

        logger.debug(f"Picked primer {picked}")
        logger.debug(f"Picked MSA slice: {picked.annotated_msa}")
        for i, primer in enumerate(candidates[:10]):
            max_mis = max(primer.mismatch_counts)
            total_mis = sum(primer.mismatch_counts)
            logger.debug(
                f"{direction.name} candidate {i}: "
                f"{'*interaction, ' if primer.interacts_with else ''}"
                f"base pen {primer.base_penalty:.3f}, "
                f"combined pen {primer.combined_penalty:.3f}, "
                f"{max_mis} max mismatch{'' if max_mis == 1 else 'es'}, "
                f"{total_mis} total mismatch{'' if total_mis == 1 else 'es'}."
            )


class NoSuitablePrimersError(Exception):
    """Unable to find suitable primers at the current cursor position."""

    pass


class SliceOutOfBoundsError(Exception):
    """The requested start position would put the slice out of bounds."""

    pass



import pytest
import random
import uuid

from pathlib import Path

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from click.testing import CliRunner


def seq_record_factory(seq_len=5000, alphabet="acgt", id=""):
    """Generate a random SeqRecord for testing purposes"""
    id = id or f"random_seq_{uuid.uuid4()}"
    seq = "".join([random.choice(alphabet) for i in range(seq_len)])
    return SeqRecord(Seq(seq), id=id)


def multi_seq_generator(n, **kwargs):
    """Generate multiple random SeqRecords for testing purposes"""
    return [seq_record_factory(**kwargs) for i in range(n)]


@pytest.fixture(scope="session")
def stored_inputs_path():
    """Return path to stored test inputs dir"""
    return Path(__file__).parent / "inputs"


@pytest.fixture(scope="session")
def temp_inputs_path(tmp_path_factory):
    """Return a temp_path for generating input files"""
    return tmp_path_factory.mktemp("inputs")


@pytest.fixture(scope="session")
def input_fasta_empty(temp_inputs_path):
    """Generate an empty fasta input file"""
    fh = temp_inputs_path / "empty_input.fa"
    fh.write_text("")
    return fh


@pytest.fixture(scope="session")
def input_fasta_invalid_alphabet(temp_inputs_path):
    """Generate a fasta file with one valid record and one with an invalid alphabet"""
    fh = temp_inputs_path / "invalid_alphabet.fa"
    records = [seq_record_factory(), seq_record_factory(alphabet="acgtRN")]
    SeqIO.write(records, fh, "fasta")
    return fh


@pytest.fixture(scope="session")
def input_fasta_5_random_valid(temp_inputs_path):
    """Generate a random multi-FASTA with 5 valid records"""
    fh = temp_inputs_path / "valid_random_5_fasta.fa"
    SeqIO.write(multi_seq_generator(5), fh, "fasta")
    return fh


@pytest.fixture(scope="session")
def input_fasta_101_random_valid(temp_inputs_path):
    """Generate a random multi-FASTA with 101 records"""
    fh = temp_inputs_path / "valid_random_101_fasta.fa"
    SeqIO.write(multi_seq_generator(101), fh, "fasta")
    return fh


@pytest.fixture(scope="session")
def input_fasta_valid_with_gaps(temp_inputs_path):
    """Generate a FASTA that includes gaps - """
    fh = temp_inputs_path / "valid_fasta_with_gaps.fa"
    SeqIO.write(seq_record_factory(alphabet="acgt-"), fh, "fasta")
    return fh


@pytest.fixture(scope="session")
def input_fasta_size_difference_over_500(temp_inputs_path):
    """Generate a FASTA where the  record is over 500nt longer than the first"""
    fh = temp_inputs_path / "invalid_different.fa"
    records = [seq_record_factory(seq_len=5000), seq_record_factory(seq_len=5501)]
    SeqIO.write(records, fh, "fasta")
    return fh


@pytest.fixture(scope="session")
def input_fasta_short_500(temp_inputs_path):
    """Generate a FASTA where the record is short"""
    fh = temp_inputs_path / "invalid_short.fa"
    SeqIO.write(seq_record_factory(seq_len=500), fh, "fasta")
    return fh


STORED_INPUTS = [
    "CHIKV_demo.fa",
    "Ebov-10-Pan.fasta",
    "nCov-2019.fasta",
    "embB_high_gc.fasta",
]


@pytest.fixture(params=STORED_INPUTS)
def all_stored_inputs(request, stored_inputs_path):
    """Return all stored input paths"""
    return stored_inputs_path / request.param


@pytest.fixture(scope="session")
def chikv_input(stored_inputs_path):
    """Return all stored input paths"""
    return stored_inputs_path / STORED_INPUTS[0]


@pytest.fixture(scope="session")
def ebola_input(stored_inputs_path):
    """Return ebola input path"""
    return stored_inputs_path / STORED_INPUTS[1]


@pytest.fixture(scope="session")
def ncov_single_ref_input(stored_inputs_path):
    """Return ebola input path"""
    return stored_inputs_path / STORED_INPUTS[2]


@pytest.fixture(scope="session")
def high_gc_input(stored_inputs_path):
    """Return ebola input path"""
    return stored_inputs_path / STORED_INPUTS[3]


@pytest.fixture(scope="session")
def random_reference_slice():
    record = seq_record_factory(seq_len=500)
    return record


@pytest.fixture(scope="session")
def cli_runner():
    return CliRunner()



"""
PrimalScheme: a primer3 wrapper for designing multiplex primer schemes

Copyright (C) 2020 Joshua Quick and Andrew Smith
www.github.com/aresti/primalscheme

This module contains classes and functions related to primer design.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>
"""

from collections import namedtuple
from enum import Enum
from itertools import groupby

from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from primer3 import calcTm as p3_calcTm, calcHairpin as p3_calcHairpin
from primalscheme import config

Kmer = namedtuple("Kmer", "seq start")


class Direction(Enum):
    """Primer direction."""

    LEFT = "+"
    RIGHT = "-"


class Primer:
    """A primer."""

    def __init__(self, seq, start, direction, pool, reference_msa):
        """Init Primer."""
        self.seq = seq
        self.start = start
        self.direction = direction
        self.pool = pool
        self.reference_msa = reference_msa
        self.interacts_with = None

    def __str__(self):
        """Primer string representation."""
        return f"{self.direction.name}:{self.seq}:{self.start}"

    @property
    def size(self):
        """Primer size (length)."""
        return len(self.seq)

    @property
    def end(self):
        """Primer (inclusive) end position."""
        if self.direction == Direction.LEFT:
            return self.start + self.size - 1
        elif self.direction == Direction.RIGHT:
            return self.start - self.size + 1

    @property
    def seq(self):
        """Primer sequence."""
        return self.__seq

    @seq.setter
    def seq(self, seq):
        """Set seq and calculate derived characteristics."""
        self.__seq = seq
        self.gc = calc_gc(self.seq)
        self.tm = calc_tm(self.seq)
        self.hairpin = calc_hairpin(self.seq)
        self.max_homo = calc_max_homo(self.seq)
        self.__base_penalty = None

    @property
    def base_penalty(self):
        """
        Calculate intrinsic primer penalty.
        As per routine described in http://primer3.ut.ee/primer3web_help.htm
        Lazy-evaluation.
        """
        if self.__base_penalty is not None:
            return self.__base_penalty

        penalty = 0

        # High Tm
        if self.tm > config.PRIMER_OPT_TM:
            penalty += config.PRIMER_WT_TM_GT * (self.tm - config.PRIMER_OPT_TM)

        # Low Tm
        if self.tm < config.PRIMER_OPT_TM:
            penalty += config.PRIMER_WT_TM_LT * (config.PRIMER_OPT_TM - self.tm)

        # High GC
        if self.gc > config.PRIMER_GC_RANGE.opt:
            penalty += config.PRIMER_WT_GC_PERCENT_GT * (
                self.gc - config.PRIMER_GC_RANGE.opt
            )

        # Low GC
        if self.gc < config.PRIMER_GC_RANGE.opt:
            penalty += config.PRIMER_WT_GC_PERCENT_LT * (
                config.PRIMER_GC_RANGE.opt - self.gc
            )

        # High size
        if self.size > config.PRIMER_SIZE_RANGE.opt:
            penalty += config.PRIMER_WT_SIZE_GT * (
                self.size - config.PRIMER_SIZE_RANGE.opt
            )

        # Low size
        if self.size < config.PRIMER_SIZE_RANGE.opt:
            penalty += config.PRIMER_WT_SIZE_LT * (
                config.PRIMER_SIZE_RANGE.opt - self.size
            )

        self.__base_penalty = penalty
        return penalty

    def _mismatch_count_for_ref(self, ref):
        """Mismatch count for the primer against a given reference."""
        count = 0
        for i, base in enumerate(self.seq):
            if base != str(ref.seq)[i]:
                count += 1
        return count

    def _mismatch_penalties_for_ref(self, ref):
        """Mismatch penalties for each position against a given reference."""
        positions = []
        penalties = config.PRIMER_MISMATCH_PENALTIES
        ref_seq = str(ref.seq)
        for i, base in enumerate(self.seq):
            if base == ref_seq[i]:
                positions.append(0)
                continue
            dist_3p = self.size - i
            if dist_3p > len(penalties) - 1:
                positions.append(penalties[-1])
            else:
                positions.append(penalties[dist_3p])
        return positions

    @property
    def mismatch_counts(self):
        """List of mismatch counts between the primer and all references."""
        return [self._mismatch_count_for_ref(ref) for ref in self.reference_msa]

    @property
    def mismatch_penalty_matrix(self):
        """Matrix of mistmatch penalties for each position, against all references."""
        return [self._mismatch_penalties_for_ref(ref) for ref in self.reference_msa]

    @property
    def mismatch_penalties(self):
        """List of total mismatch penalties for the primer against all references."""
        return [sum(row) for row in self.mismatch_penalty_matrix]

    @property
    def combined_penalty(self):
        """Combined penalty (base + mistmatch)."""
        return self.base_penalty + sum(self.mismatch_penalties)

    @property
    def reference_msa(self):
        """Reference MSA, sliced to match primer position."""
        return self.__reference_msa

    @reference_msa.setter
    def reference_msa(self, msa):
        """Set reference MSA."""
        self.__reference_msa = msa
        self.__annotated_msa = None

    @property
    def annotated_msa(self):
        """Reference MSA with annotated cigar for logging (lazy eval)."""
        if self.__annotated_msa:
            return self.__annotated_msa

        # Calculate cigar
        cigar = ""
        for i in range(self.__reference_msa.get_alignment_length()):
            num = len(set(self.__reference_msa[::-1, i]))
            if num == 1:
                cigar += "*"
            else:
                cigar += " "
        cigar = SeqRecord(Seq(cigar), id="Cigar")
        self.__annotated_msa = self.__reference_msa[:]
        self.__annotated_msa.append(cigar)
        return self.__annotated_msa


def calc_gc(seq):
    """Calculate percent GC for a sequence."""
    return 100.0 * (seq.count("G") + seq.count("C")) / len(seq)


def calc_tm(seq):
    """Calculate Tm for a sequence."""
    return p3_calcTm(
        seq,
        mv_conc=config.MV_CONC,
        dv_conc=config.DV_CONC,
        dntp_conc=config.DNTP_CONC,
        dna_conc=config.DNA_CONC,
    )


def calc_max_homo(seq):
    """Calculate max homopolymer length for a sequence."""
    return sorted([(len(list(g))) for k, g in groupby(seq)], reverse=True)[0]


def calc_hairpin(seq):
    """Calculate hairpin Tm for a sequence."""
    return p3_calcHairpin(
        seq,
        mv_conc=config.MV_CONC,
        dv_conc=config.DV_CONC,
        dntp_conc=config.DNTP_CONC,
        dna_conc=config.DNA_CONC,
    ).tm


def design_primers(msa, direction, pool, offset=0, primary_only=False):
    """Design primers against a reference MSA."""
    min_size = config.PRIMER_SIZE_RANGE.min
    max_size = config.PRIMER_SIZE_RANGE.max
    variation = max_size - min_size

    if primary_only:
        sequences = [msa[0].seq]
    else:
        sequences = [record.seq for record in msa]

    # Digest all sequences into k-mers
    all_kmers = set()
    for seq in sequences:
        for kmer_size in range(min_size, max_size + 1):
            all_kmers.update(digest_seq(str(seq[:-variation]), kmer_size))

    # Generate primers
    primers = []
    for kmer in all_kmers:
        sliced_msa = msa[:, kmer.start : kmer.start + len(kmer.seq)]
        if direction == Direction.LEFT:
            primer_start = offset + kmer.start
        else:
            primer_start = offset + len(msa[0].seq) - 1 - kmer.start
        primers.append(Primer(kmer.seq, primer_start, direction, pool, sliced_msa))

    return primers


def primer_thermo_filter(primer):
    """Hard filter for candidate primers."""
    return (
        (config.PRIMER_GC_RANGE.min <= primer.gc <= config.PRIMER_GC_RANGE.max)
        and (config.PRIMER_MIN_TM <= primer.tm <= config.PRIMER_MAX_TM)
        and (primer.hairpin <= config.PRIMER_MAX_HAIRPIN_TH)
        and (primer.max_homo <= config.PRIMER_MAX_HOMO)
    )


def digest_seq(seq, kmer_size):
    """Digest a sequence into kmers."""
    return [Kmer(seq[i : i + kmer_size], i) for i in range((len(seq) - kmer_size) + 1)]



"""
Tests for command line interface (CLI)
"""

import os
import pytest
import re
import warnings

from Bio.SeqRecord import SeqRecord
from primalscheme.cli import cli

with warnings.catch_warnings():
    warnings.simplefilter("ignore", PendingDeprecationWarning)
    from primalscheme.cli import process_fasta


def test_runas_module():
    """Can this package be run as a Python module?"""
    exit_status = os.system("python -m primalscheme --help")

    assert exit_status == 0


def test_entrypoint():
    """Is entrypoint script installed?"""
    exit_status = os.system("primalscheme --help")

    assert exit_status == 0


def test_mutliplex_command(cli_runner):
    """Is multiplex command available, with help?"""
    result = cli_runner.invoke(cli, ["multiplex", "-h"])

    assert result.exit_code == 0


def test_cli_without_command_shows_help(cli_runner):
    """Does CLI display help when run w/o a command argument?"""
    result = cli_runner.invoke(cli)

    assert "--help" in result.output


def test_multiplex_command_without_fasta_returns_exit_code_2(cli_runner):
    """Does CLI stop execution w/o a positional fasta arg, returning exit code 2?"""
    result = cli_runner.invoke(cli, ["multiplex"])

    assert result.exit_code == 2


@pytest.mark.parametrize(
    "option, value",
    [
        ("--amplicon-size", "380"),
        ("-a", "420"),
        ("--outpath", "./scheme"),
        ("-o", "./scheme"),
        ("--name", "myScheme"),
        ("-n", "myScheme"),
        ("--target-overlap", "0"),
        ("-t", "0"),
        ("--debug", None),
        ("-d", None),
        ("--force", None),
        ("-f", None),
        ("--pinned", None),
        ("-p", None),
        ("--high-gc", None),
        ("-g", None),
        ("--help", None),
        ("-h", None),
    ],
)
def test_availability_of_cli_options(option, value, cli_runner):
    """Are options available?"""
    args = ["multiplex", "-h", option]
    if value:
        args.append(value)
    result = cli_runner.invoke(cli, args)

    assert "Error: no such option" not in result.output


def test_cli_fails_with_invalid_option(chikv_input, cli_runner):
    """Does CLI stop execution with an invalid option?"""
    args = ["multiplex", f"{str(chikv_input)}", "--force", "--badOption"]
    result = cli_runner.invoke(cli, args)

    assert result.exit_code == 2
    assert "Error: no such option" in result.output


@pytest.mark.parametrize(
    "option", ["-V", "--version"],
)
def test_cli_version_output(option, cli_runner):
    """Does CLI output a sensible version number for -V and --version?"""
    result = cli_runner.invoke(cli, option)

    assert re.match(
        r"^cli, version \d{1,2}.\d{1,2}.\d{1,2}[a-zA-Z]{0,3}\d*$", result.output,
    )


def test_force_not_required_when_outpath_does_not_exist(
    tmp_path, cli_runner, chikv_input
):
    """Does CLI proceed without --force when outpath does not exist?"""
    path = str(tmp_path / "output")
    result = cli_runner.invoke(cli, ["multiplex", str(chikv_input), "-o", path, "-p"])

    assert result.exit_code == 0


def test_force_required_when_output_path_does_exist(tmp_path, cli_runner, chikv_input):
    """Does CLI require --force when outpath does exist?"""
    path = tmp_path / "output"
    path.mkdir()
    result = cli_runner.invoke(
        cli, ["multiplex", str(chikv_input), "-o", str(path), "-p"]
    )

    assert result.exit_code != 0
    assert "--force" in result.output


def test_force_allows_existing_output_path(tmp_path, cli_runner, chikv_input):
    """Does CLI proceed with an existing outpath when --force is set?"""
    path = tmp_path / "output"
    path.mkdir()
    result = cli_runner.invoke(
        cli, ["multiplex", str(chikv_input), "-o", str(path), "-p", "-f"]
    )

    assert result.exit_code == 0


def test_existing_output_path_not_dir(tmp_path, cli_runner, chikv_input):
    """Does CLI require outpath to be a directory"""
    path = tmp_path / "output"
    path.write_text("text")
    result = cli_runner.invoke(
        cli, ["multiplex", str(chikv_input), "-o", str(path), "-p"]
    )

    assert result.exit_code != 0
    assert "file" in result.output


def test_warning_on_high_gc_input(tmp_path, cli_runner, high_gc_input):
    """Does CLI warn user when using a high-GC input without the --high-gc option?"""
    path = tmp_path / "output"
    result = cli_runner.invoke(cli, ["multiplex", str(high_gc_input), "-o", str(path)])

    assert "warning:" in result.output.lower()
    assert "high-gc" in result.output.lower()


def test_warning_silenced_when_using_high_gc(tmp_path, cli_runner, high_gc_input):
    """Does CLI silence the warning when using the --high-gc option?"""
    path = tmp_path / "output"
    result = cli_runner.invoke(
        cli, ["multiplex", str(high_gc_input), "-o", str(path), "-g"]
    )

    assert "warning:" not in result.output.lower()
    assert "high-gc" not in result.output.lower()


def test_no_high_gc_warning_on_average_input(
    tmp_path, cli_runner, ncov_single_ref_input
):
    """Does CLI stay quiet when using a regular input wihout the --high-gc option?"""
    path = tmp_path / "output"
    result = cli_runner.invoke(
        cli, ["multiplex", str(ncov_single_ref_input), "-o", str(path)]
    )

    assert "warning:" not in result.output.lower()
    assert "high-gc" not in result.output.lower()


def test_too_short_fasta_size_vs_amplicon_size(
    input_fasta_short_500, tmp_path, cli_runner
):
    """Does CLI exit where amplicon size is greater than reference size?"""
    path = str(tmp_path / "output")
    args = [
        "multiplex",
        str(input_fasta_short_500),
        "-a",
        "500",
        "-a",
        "550",
        "-o",
        path,
    ]
    result = cli_runner.invoke(cli, args)

    assert result.exit_code != 0
    assert "too short" in result.output


def test_process_fasta_empty_input(input_fasta_empty):
    """Does process_fasta raise for empty input?"""
    with pytest.raises(ValueError, match="does not contain any valid references"):
        process_fasta(input_fasta_empty)


def test_process_fasta_invalid_alphabet(input_fasta_invalid_alphabet):
    """Does process_fasta raise for invalid alphabet?"""
    with pytest.raises(ValueError, match="invalid nucleotide codes"):
        process_fasta(input_fasta_invalid_alphabet)


def test_process_fasta_too_many_records(input_fasta_101_random_valid):
    """Does process_fasta raise for too many references?"""
    with pytest.raises(ValueError, match="A maximum of 100"):
        process_fasta(input_fasta_101_random_valid)


def test_process_fasta_size_difference_over_500(input_fasta_size_difference_over_500,):
    """Does process_fasta raise for size difference over 500?"""
    with pytest.raises(ValueError, match="too different in size"):
        process_fasta(input_fasta_size_difference_over_500)


def test_process_fasta_returns_list_of_seq_records(input_fasta_5_random_valid):
    """Does process_fasta return a list of SeqRecord objects?"""
    references = process_fasta(input_fasta_5_random_valid)
    assert len(references) == 5
    assert all(isinstance(r, SeqRecord) for r in references)


def test_process_fasta_chikv_demo(chikv_input):
    """Does process_fasta return 2 SeqRecord objects for demo CHIKV input?"""
    references = process_fasta(chikv_input)
    assert len(references) == 2


def test_process_fasta_gaps_removed(input_fasta_valid_with_gaps):
    """Does process_fasta remove gaps?"""
    references = process_fasta(input_fasta_valid_with_gaps)
    assert "-" not in references[0]


def test_process_fasta_too_short_input(input_fasta_short_500):
    """Does process_fasta raise for too short input?"""
    with pytest.raises(ValueError, match="too short"):
        process_fasta(input_fasta_short_500, min_ref_size=900)



"""
PrimalScheme: a primer3 wrapper for designing multiplex primer schemes

Copyright (C) 2020 Joshua Quick and Andrew Smith
www.github.com/aresti/primalscheme

This module contains the MutilplexReporter object.
This object extends a MultiplexScheme object to provide reporting methods.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>
"""

import csv
import json
import logging
import pickle

from abc import ABC, abstractmethod
from collections import namedtuple

from Bio import SeqIO
from Bio.Graphics import GenomeDiagram
from Bio.SeqFeature import FeatureLocation, SeqFeature
from reportlab.lib import colors

from primalscheme import config, __version__ as version
from primalscheme.multiplex import MultiplexScheme
from primalscheme.primer import Direction

logger = logging.getLogger("primalscheme")


Insert = namedtuple("Insert", "start end pool")


class MultiplexReporter(MultiplexScheme):
    """Reporting methods to extend MultiplexScheme."""

    def __init__(self, outpath, *args, **kwargs):
        """Init MultiplexReporter."""
        self.outpath = outpath
        super().__init__(*args, **kwargs)

    @property
    def inserts(self):
        """A list of insert (start, end) tuples."""
        return [Insert(r.left.end + 1, r.right.end - 1, r.pool) for r in self.regions]

    @property
    def gap_count(self):
        """The total number of gaps in the scheme."""
        gaps = 0
        last_covered = self.inserts[0].end
        for insert in self.inserts[1:]:
            if insert.start > last_covered:
                gaps += 1
            last_covered = insert.end
        return gaps

    @property
    def percent_coverage(self):
        """Coverage %, with respect to the primary reference."""
        covered_coords = set(
            [x for insert in self.inserts for x in range(insert.start, insert.end + 1)]
        )
        return round(len(covered_coords) / self.ref_len * 100, 2)

    def calc_gc(self, sequence):
        """Return gc content for a sequence as fraction."""
        g = sequence.count("G") + sequence.count("g")
        c = sequence.count("C") + sequence.count("c")
        return (g + c) / len(sequence)

    def write_default_outputs(self):
        """Write all default output files."""
        self.write_primer_bed()
        self.write_insert_bed()
        self.write_primer_tsv()
        self.write_refs()
        self.write_schemadelica_plot()
        self.write_run_report_json()

    def write_primer_bed(self):
        """Write primer BED file."""
        filepath = self.outpath / f"{self.prefix}.primer.bed"
        logger.info(f"Writing {filepath}")

        rows = []
        ref_id = self.primary_ref.id

        for i, p in enumerate(self.primers):
            region_num = int(i / 2) + 1
            start = p.start if p.direction == Direction.LEFT else p.end
            end = p.end if p.direction == Direction.LEFT else p.start
            name = f"{self.prefix}_{region_num}_{p.direction.name}"
            rows.append(
                [
                    ref_id,
                    start,
                    end + 1,  # BED end is 1-based
                    name,
                    p.pool,
                    p.direction.value,
                ]
            )

        with open(filepath, "w") as fh:
            cw = csv.writer(fh, delimiter="\t")
            cw.writerows(rows)

    def write_insert_bed(self):
        """Write insert BED file."""
        filepath = self.outpath / f"{self.prefix}.insert.bed"
        logger.info(f"Writing {filepath}")

        rows = []
        ref_id = self.primary_ref.id

        for insert_num, insert in enumerate(self.inserts):
            rows.append(
                [
                    ref_id,
                    insert.start,
                    insert.end + 1,  # BED end is 1-based
                    f"{self.prefix}_INSERT_{insert_num + 1}",
                    insert.pool,
                    "+",
                ]
            )

        with open(filepath, "w") as fh:
            cw = csv.writer(fh, delimiter="\t")
            cw.writerows(rows)

    def write_primer_tsv(self):
        """Write primer TSV file."""
        filepath = self.outpath / f"{self.prefix}.primer.tsv"
        logger.info(f"Writing {filepath}")

        rows = [["name", "pool", "seq", "size", "%gc", "tm (use 65)"]]

        for i, p in enumerate(self.primers):
            region_num = int(i / 2) + 1
            rows.append(
                [
                    f"{self.prefix}_{region_num}_{p.direction.name}",
                    p.pool,
                    p.seq,
                    p.size,
                    f"{p.gc:.2f}",
                    f"{p.tm:.2f}",
                ]
            )

        with open(filepath, "w") as fh:
            cw = csv.writer(fh, delimiter="\t")
            cw.writerows(rows)

    def write_pickle(self):
        """Write pickle file."""
        filepath = self.outpath / f"{self.prefix}.pickle"
        logger.info(f"Writing {filepath}")
        with open(filepath, "wb") as pickleobj:
            pickle.dump(self, pickleobj)

    def write_refs(self):
        """Write reference FASTA."""
        filepath = self.outpath / f"{self.prefix}.reference.fasta"
        logger.info(f"Writing {filepath}")
        with open(filepath, "w"):
            SeqIO.write(self.references, filepath, "fasta")

    def apply_to_window(self, sequence, window_size, function, step=None):
        """
        Modified from
        https://github.com/biopython/biopython/blob/master/Tests/test_GenomeDiagram.py
        Apply function to windows of the given sequence.
        Returns a list of (position, value) tuples for fragments of the passed
        sequence of length window_size (stepped by step), calculated by the
        passed function.  Returned positions are the midpoint of each window.
        Arguments:
        - sequence - Bio.Seq.Seq object.
        - window_size - an integer describing the length of sequence
          to consider.
        - step - an integer describing the step to take between windows
          (default = window_size//2).
        - function - Method or function that accepts a Bio.Seq.Seq object
          as its sole argument and returns a single value.
        apply_to_window(sequence, window_size, function) ->
            [(int, float),(int, float),...]
        """
        seqlen = len(sequence)  # Total length of sequence to be used
        if step is None:  # Use half window-width or 1 if larger
            step = max(window_size // 2, 1)
        else:  # Use specified step, or 1 if greater
            step = max(step, 1)

        results = []  # Holds (position, value) results

        # Perform the passed function on as many windows as possible, short of
        # overrunning the sequence
        pos = 0
        while pos < seqlen - window_size + 1:
            # Obtain sequence fragment
            start = pos
            middle = (pos + window_size + pos) // 2
            end = pos + window_size
            fragment = sequence[start:end]
            # Apply function to the sequence fragment
            value = function(fragment)
            results.append((middle, value))  # Add results to list
            # Advance to next fragment
            pos += step

        # Use the last available window on the sequence, even if it means
        # re-covering old ground
        if pos != seqlen - window_size:
            # Obtain sequence fragment
            pos = seqlen - window_size
            start = pos
            middle = (pos + window_size + pos) // 2
            end = pos + window_size
            fragment = sequence[start:end]
            # Apply function to sequence fragment
            value = function(fragment)
            results.append((middle, value))  # Add results to list

        return results  # Return the list of (position, value) results

    def write_schemadelica_plot(self):
        """Write schemadelica plot as SVG and PDF."""
        gd_diagram = GenomeDiagram.Diagram("Primer Scheme", track_size=0.15)
        primer_feature_set = GenomeDiagram.FeatureSet()

        # make the gc track
        window = 50
        gc_set = GenomeDiagram.GraphSet("GC content")
        graphdata1 = self.apply_to_window(self.primary_ref.seq, window, self.calc_gc)
        gc_set.new_graph(
            graphdata1,
            "GC content",
            style="line",
            color=colors.violet,
            altcolor=colors.purple,
        )
        gc_track = GenomeDiagram.Track(
            "GC content", height=1.5, greytrack=0, scale_largetick_interval=1e3
        )
        gc_track.add_set(gc_set)

        # make the primer track
        for r in self.regions:
            region = str(r.region_num)
            strand = 1 if r.region_num % 2 else -1

            fwd_feature = SeqFeature(
                FeatureLocation(r.left.start, r.left.end, strand=strand)
            )
            rev_feature = SeqFeature(
                FeatureLocation(r.right.end, r.right.start, strand=strand)
            )
            region_feature = SeqFeature(
                FeatureLocation(r.left.start, r.right.start, strand=strand)
            )

            primer_color = colors.red
            region_color = colors.palevioletred

            primer_feature_set.add_feature(
                region_feature,
                color=region_color,
                name=region,
                label=True,
                label_position="middle",
                label_angle=0 if strand == 1 else -180,
            )
            primer_feature_set.add_feature(fwd_feature, color=primer_color, name=region)
            primer_feature_set.add_feature(rev_feature, color=primer_color, name=region)

        primer_track = GenomeDiagram.Track(name="Annotated Features", height=1)
        primer_track.add_set(primer_feature_set)

        gd_diagram.add_track(primer_track, 2)
        gd_diagram.add_track(gc_track, 1)

        rows = max(2, int(round(len(self.primary_ref) / 10000.0)))
        gd_diagram.draw(
            format="linear",
            pagesize=(300 * rows, 200 * rows),
            fragments=rows,
            start=0,
            end=len(self.primary_ref),
        )

        pdf_filepath = self.outpath / f"{self.prefix}.plot.pdf"
        svg_filepath = self.outpath / f"{self.prefix}.plot.svg"
        logger.info(f"Writing {pdf_filepath}")
        logger.info(f"Writing {svg_filepath}")
        gd_diagram.write(str(pdf_filepath), "PDF", dpi=300)
        gd_diagram.write(str(svg_filepath), "SVG", dpi=300)

    def write_run_report_json(self):
        """Write run report json."""
        filepath = self.outpath / f"{self.prefix}.report.json"
        logger.info(f"Writing {filepath}")
        data = {
            "references": [ref.id for ref in self.references],
            "primary_ref": self.references[0].id,
            "secondary_refs": [ref.id for ref in self.secondary_refs],
            "secondary_ref_count": len(self.secondary_refs),
            "excluded_refs": [ref.id for ref in self.excluded_refs],
            "excluded_ref_count": len(self.excluded_refs),
            "regions": len(self.regions),
            "percent_coverage": self.percent_coverage,
            "gaps": self.gap_count,
            "config": {
                "amplicon_size_min": self.amplicon_size_min,
                "amplicon_size_max": self.amplicon_size_max,
                "target_overlap": self.target_overlap,
                "high_gc": self.high_gc,
                "step_distance": config.STEP_DISTANCE,
                "primer_size_range": {
                    "min": config.PRIMER_SIZE_RANGE.min,
                    "max": config.PRIMER_SIZE_RANGE.max,
                    "opt": config.PRIMER_SIZE_RANGE.opt,
                },
                "primer_gc_range": {
                    "min": config.PRIMER_GC_RANGE.min,
                    "max": config.PRIMER_GC_RANGE.max,
                    "opt": config.PRIMER_GC_RANGE.opt,
                },
                "primalscheme_version": version,
            },
        }
        filepath.write_text(json.dumps(data))


class ProgressTracker(ABC):
    """Abstract base class for ProgressTracker."""

    @abstractmethod
    def goto(self, val):
        """Update progress to val."""
        ...

    @property
    def end(self):
        """Progress end value."""
        ...

    @end.setter
    @abstractmethod
    def end(self, val):
        """Set progress end value."""
        ...

    @property
    def considered(self):
        """Count of considered primers."""
        ...

    @considered.setter
    @abstractmethod
    def considered(self, considered):
        """Set count of considered primers."""
        ...

    @property
    def region_num(self):
        """The current region num."""
        ...

    @region_num.setter
    @abstractmethod
    def region_num(self, region_num):
        """Set current region num."""
        ...

    @abstractmethod
    def interrupt(self):
        """Prepare to be interrupted by a log message."""
        ...

    @abstractmethod
    def finish(self):
        """Finish tracking progress."""
        ...



"""
PrimalScheme: a primer3 wrapper for designing multiplex primer schemes

Copyright (C) 2020 Joshua Quick and Andrew Smith
www.github.com/aresti/primalscheme

This module contains config values.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>
"""

from collections import namedtuple

PrimerSizeRange = namedtuple("PrimerSizeRange", "min max opt")
PRIMER_SIZE_RANGES = {
    "DEFAULT": PrimerSizeRange(19, 34, 22),
    "HIGH_GC": PrimerSizeRange(17, 30, 20),
}

PrimerGCRange = namedtuple("GCRange", "min max opt")
PRIMER_GC_RANGES = {
    "DEFAULT": PrimerGCRange(30, 55, 50),
    "HIGH_GC": PrimerGCRange(40, 65, 60),
}

PREFIX = "scheme"
OUTPUT_PATH = "./output"
MAX_ALN_GAP_PERCENT = 0.03

TARGET_OVERLAP = 0
STEP_DISTANCE = 11
AMPLICON_SIZE_MIN = 380
AMPLICON_SIZE_MAX = 420
SIZE_RANGE_AUTO = 0.1
HIGH_GC_WARN_THRESHOLD = 55

PRIMER_SIZE_RANGE = PRIMER_SIZE_RANGES["DEFAULT"]
PRIMER_OPT_TM = 61.0
PRIMER_WT_TM_GT = 1.0
PRIMER_WT_TM_LT = 1.0
PRIMER_MIN_TM = 59.5
PRIMER_MAX_TM = 62.5
PRIMER_WT_SIZE_GT = 1.0
PRIMER_WT_SIZE_LT = 1.0
PRIMER_WT_GC_PERCENT_GT = 0.0
PRIMER_WT_GC_PERCENT_LT = 0.0
PRIMER_GC_RANGE = PRIMER_GC_RANGES["DEFAULT"]
PRIMER_MAX_HAIRPIN_TH = 50
PRIMER_MAX_HOMO = 5

PRIMER_MAX_MISMATCHES = 2
PRIMER_MISMATCH_PENALTIES = [3, 2, 1]  # distance from 3', last is default

MV_CONC = 100.0
DV_CONC = 2.0
DNA_CONC = 15.0
DNTP_CONC = 0.8
TEMP_C = 65.0
HETERODIMER_DG_THRESHOLD = -2.0



"""
PrimalScheme: a primer3 wrapper for designing multiplex primer schemes

Copyright (C) 2020 Joshua Quick and Andrew Smith
www.github.com/aresti/primalscheme

This module contains the CLI for PrimalScheme.
It is executed when the user runs 'primalscheme' after installation,
or 'python -m primalscheme'.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>
"""

import click
import logging
import sys

from pathlib import Path

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from progress.bar import ShadyBar

from primalscheme import __version__ as version, config
from primalscheme.primer import calc_gc
from primalscheme.region import NoSuitablePrimersError
from primalscheme.reporting import MultiplexReporter, ProgressTracker

logger = logging.getLogger("primalscheme")


CLI_CONTEXT = dict(auto_envvar_prefix="PRIMAL", help_option_names=["-h", "--help"],)


@click.group(context_settings=CLI_CONTEXT)
@click.version_option(version, "--version", "-V")
def cli():
    """a tool for designing primer panels for multiplex PCR."""
    pass


@cli.command()
@click.argument("fasta", type=click.Path(exists=True, dir_okay=False))
@click.option(
    "--amplicon-size",
    "-a",
    multiple=True,
    type=click.IntRange(90),
    help=(
        "Amplicon size target. Pass twice to set an exact range, otherwise expect "
        f"+/- {config.SIZE_RANGE_AUTO/2 * 100}%."
    ),
    metavar="<int>",
    default=[config.AMPLICON_SIZE_MIN, config.AMPLICON_SIZE_MAX],
    show_default=True,
)
@click.option(
    "--outpath",
    "-o",
    type=click.Path(file_okay=False, writable=True),
    help="Path to output directory.",
    metavar="<dir>",
    default=config.OUTPUT_PATH,
    show_default=True,
)
@click.option(
    "--name",
    "-n",
    type=click.STRING,
    help="Prefix name for your outputs.",
    metavar="<str>",
    default=config.PREFIX,
    show_default=True,
)
@click.option(
    "--target-overlap",
    "-t",
    type=click.IntRange(0),
    help="Target insert overlap size.",
    metavar="<int>",
    default=config.TARGET_OVERLAP,
    show_default=True,
)
@click.option("--debug/--no-debug", "-d", help="Set log level DEBUG.", default=False)
@click.option(
    "--force/--no-force",
    "-f",
    help="Force output to an existing directory, overwrite files.",
    default=False,
)
@click.option(
    "--pinned/--no-pinned",
    "-p",
    help="Only consider primers from the first reference.",
    default=False,
)
@click.option(
    "--high-gc/--no-high-hc",
    "-g",
    help="Use config suitable for high-GC sequences.",
    default=False,
)
def multiplex(
    fasta, amplicon_size, outpath, name, target_overlap, debug, force, pinned, high_gc
):
    """Design a multiplex PCR scheme."""
    # Handle args
    if len(amplicon_size) == 1:
        target_amplicon_size = amplicon_size[0]
        half_range = int(target_amplicon_size * config.SIZE_RANGE_AUTO / 2)
        amplicon_size_min = target_amplicon_size - half_range
        amplicon_size_max = target_amplicon_size + half_range
    else:
        amplicon_size_min = min(amplicon_size[:2])
        amplicon_size_max = max(amplicon_size[:2])

    # Validate output path
    try:
        outpath = get_output_path(outpath, force=force)
    except IOError as e:
        click.echo(click.style(f"Error: {e}", fg="red",))
        sys.exit(1)

    # Setup logging
    setup_logging(outpath, debug=debug, prefix=name)

    # Process FASTA input
    try:
        references = process_fasta(fasta, min_ref_size=amplicon_size_max)
    except ValueError as e:
        logger.error(f"Error: {e}")
        sys.exit(2)

    # High GC warning
    if not high_gc and any(
        calc_gc(str(ref.seq)) > config.HIGH_GC_WARN_THRESHOLD for ref in references
    ):
        click.echo(
            click.style(
                "WARNING: High-GC reference detected. Consider using --high-gc mode.",
                fg="red",
            )
        )

    # Progress bar
    progress_bar = ProgressBar()

    # Log references
    ref_ids = [f" - {ref.id}" for ref in references]
    logger.info("\n".join(["References:"] + ref_ids))
    logger.info(f"Primary reference for coordinate system: {references[0].id}")
    logger.info(
        "Considering primers from "
        f"{'primary reference only' if pinned else 'all references'}"
    )

    # Create scheme
    try:
        scheme = MultiplexReporter(
            outpath,
            references,
            prefix=name,
            amplicon_size_min=amplicon_size_min,
            amplicon_size_max=amplicon_size_max,
            target_overlap=target_overlap,
            primary_only=pinned,
            high_gc=high_gc,
            progress_tracker=progress_bar,
        )
        scheme.design_scheme()
    except NoSuitablePrimersError:
        # Unable to find primers for at least one region
        logger.error("Error: Unable to find suitable primers")
        sys.exit(10)
    except Exception as e:
        # Unexpected error
        logger.error(f"Error: {e}")
        raise e

    # Write outputs
    logger.info(
        f"All done! Scheme created with {len(scheme.regions)} regions, "
        f"{scheme.gap_count } gap{'' if scheme.gap_count == 1 else 's'}, "
        f"{scheme.percent_coverage}% coverage"
    )
    scheme.write_default_outputs()
    if debug:
        scheme.write_pickle()
    sys.exit(0)


def process_fasta(file_path, min_ref_size=None):
    """Parse and validate the fasta file."""

    references = []
    records = SeqIO.parse(file_path, "fasta")  # may raise

    # Remove gaps
    for record in records:
        ref = SeqRecord(
            Seq(str(record.seq).replace("-", "").upper()),
            id=record.id,
            description=record.id,
        )
        references.append(ref)

    # Check for no references
    if not references:
        raise ValueError("The input FASTA file does not contain any valid references.")

    # Check for too short references
    if min_ref_size and any(len(ref) < min_ref_size for ref in references):
        raise ValueError(
            "One or more of your references is too short. Based on your target "
            f"amplicon size, the minimum reference size is {min_ref_size} nt."
        )

    # Check for too many references
    if len(references) > 100:
        raise ValueError("A maximum of 100 reference genomes is currently supported.")

    # Check for max difference in size between references
    primary_ref = references[0]
    primary_ref_len = len(primary_ref)

    if any(abs(len(r) - primary_ref_len) > 500 for r in references):
        raise ValueError(
            "One or more of your references is too different in size to "
            "the primary (first) reference. The maximum difference is "
            "500 nt."
        )

    # Check for a valid alphabet
    VALID_ALPHABET = "AGCTacgt"
    for r in references:
        if any(base not in VALID_ALPHABET for base in set(r.seq)):
            raise ValueError(
                "One or more of your fasta sequences contain invalid "
                f"nucleotide codes. The supported alphabet is '{VALID_ALPHABET}'. "
                "Ambiguity codes and gaps are not currently supported."
            )

    return references


def setup_logging(output_path, debug=False, prefix="primalscheme"):
    """Setup logging output and verbosity."""

    logger.setLevel(logging.DEBUG if debug else logging.INFO)

    # File handler
    log_filepath = output_path / f"{prefix}.log"
    fh = logging.FileHandler(log_filepath)
    fh.setLevel(logging.DEBUG)
    fh_formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    fh.setFormatter(fh_formatter)
    logger.addHandler(fh)

    # Stream handler STDOUT
    sh = logging.StreamHandler(sys.stdout)
    sh.setLevel(logging.INFO)
    sh_formatter = logging.Formatter("%(message)s")
    sh.setFormatter(sh_formatter)
    logger.addHandler(sh)

    logger.info(f"Writing log to {log_filepath}")
    logger.debug(f"primalscheme version {version}")


def get_output_path(output_path, force=False):
    """
    Check for an existing output dir, require --force to overwrite.
    Create dir, return path object.
    """
    path = Path(output_path)

    if path.exists() and not force:
        raise IOError("Directory exists add --force to overwrite")

    path.mkdir(exist_ok=True)
    return path


class ProgressBar(ShadyBar, ProgressTracker):
    """Progress bar for terminal stdout."""

    suffix = "%(percent)d%% [%(index)d / %(max)d]"

    def __init__(self, *args, **kwargs):
        """Init ProgressBar."""
        self.__considered = 0
        self.__region_num = 1
        super().__init__(*args, **kwargs)

    def goto(self, val):
        """Update progress to val."""
        super().goto(val)

    @property
    def end(self):
        """End (max) progress value."""
        return self.max

    @end.setter
    def end(self, val):
        """Set end progress value."""
        self.max = val

    @property
    def considered(self):
        """Count of considered primers."""
        return self.__considered

    @considered.setter
    def considered(self, considered):
        """Set count of considered primers."""
        self.__considered = considered
        self.update_message()

    @property
    def region_num(self):
        """Current region num."""
        return self.__region_num

    @region_num.setter
    def region_num(self, region_num):
        """Set current region num."""
        self.__region_num = region_num
        self.update_message()

    def update_message(self):
        """Update progress bar prefix message."""
        self.message = f"Considered {self.considered} primers, region {self.region_num}"

    def interrupt(self):
        """Prepare to be interrupted by a log message."""
        if self.index:
            self.finish()


if __name__ == "__main__":
    cli()



import pytest

from Bio import SeqIO

from primalscheme.cli import process_fasta
from primalscheme.multiplex import MultiplexScheme
from primalscheme.primer import Direction


@pytest.fixture(scope="session")
def chikv_scheme(chikv_input):
    return get_scheme(chikv_input)


@pytest.fixture(scope="session")
def ebola_scheme(ebola_input):
    return get_scheme(ebola_input)


def get_scheme(fasta, **kwargs):
    references = process_fasta(fasta)
    scheme = MultiplexScheme(references, **kwargs)
    scheme.design_scheme()
    return scheme


def no_collisions(regions):
    pool_ids = (1, 2)
    for pool_id in pool_ids:
        pool = [r for r in regions if r.pool == pool_id]
        last_coord = -1
        for r in pool:
            if r.left.start <= last_coord:
                return False
            last_coord = r.right.start
    return True


def test_chikv_scheme_has_no_gaps(chikv_scheme):
    regions = chikv_scheme.regions
    inserts = [(r.left.end + 1, r.right.end) for r in regions]
    covered_coords = set([x for insert in inserts for x in range(*insert)])
    all_coords = set(range(regions[0].left.end + 1, regions[-1].right.end))

    assert all_coords.issubset(covered_coords)


def test_no_collisions_in_any_test_scheme(all_stored_inputs):
    scheme = get_scheme(all_stored_inputs)
    assert no_collisions(scheme.regions)


@pytest.mark.parametrize("amplicon_size", range(200, 801, 200))
def test_scheme_varying_amplicon_sizes(amplicon_size, chikv_input):
    variation = int(0.1 * amplicon_size / 2)
    amplicon_size_min = amplicon_size - variation
    amplicon_size_max = amplicon_size + variation
    scheme = get_scheme(
        chikv_input,
        amplicon_size_min=amplicon_size_min,
        amplicon_size_max=amplicon_size_max,
    )

    regions = scheme.regions
    for region in regions[:4]:
        product_size = region.product_size
        assert product_size <= amplicon_size_max
        assert product_size >= amplicon_size_min


def test_large_target_overlap_does_not_result_in_collision(chikv_input):
    scheme = get_scheme(chikv_input, target_overlap=300)

    assert no_collisions(scheme.regions)


def test_candidates_are_sorted_first_by_penalty(chikv_scheme):
    regions = chikv_scheme.regions
    for region in regions:
        candidates = region.left_candidates + region.right_candidates
        prev = None
        direction = "LEFT"
        for candidate in candidates:
            if candidate.direction.name != direction:
                direction = "RIGHT"
                prev = None
            if prev:
                if candidate.combined_penalty < prev:
                    pytest.fail("Candidates are not sorted first by combined_penalty.")
            prev = candidate.combined_penalty


def test_candidates_are_sorted_second_by_start(chikv_scheme):
    regions = chikv_scheme.regions
    for region in regions:
        candidates = region.left_candidates + region.right_candidates
        prev = None
        direction = "LEFT"
        for candidate in candidates:
            attrs = (candidate.combined_penalty, candidate.start)
            if candidate.direction.name != direction:
                direction = "RIGHT"
                prev = None
            if prev:
                if attrs[0] == prev[0] and attrs[1] > prev[1]:
                    pytest.fail("Candidates are not sorted second by start position.")
            prev = attrs


def test_candidates_are_sorted_third_by_seq(ebola_scheme):
    regions = ebola_scheme.regions
    for region in regions:
        candidates = region.left_candidates + region.right_candidates
        prev = None
        direction = "LEFT"
        for candidate in candidates:
            attrs = (candidate.combined_penalty, candidate.start, candidate.seq)
            if candidate.direction.name != direction:
                direction = "RIGHT"
                prev = None
            if prev:
                if attrs[0] == prev[0] and attrs[1] == prev[1] and attrs[2] < prev[2]:
                    pytest.fail("Candidates are not sorted third by seq.")
            prev = attrs


def test_left_primer_seq_matches_some_ref_slice(chikv_scheme):
    left = chikv_scheme.regions[0].left
    ref_slices = left.reference_msa

    assert any(ref_slice.seq == left.seq for ref_slice in ref_slices)


def test_right_primer_seq_matches_some_ref_slice(chikv_scheme):
    right = chikv_scheme.regions[0].right
    ref_slices = right.reference_msa

    assert any(ref_slice.seq == right.seq for ref_slice in ref_slices)


def test_multiple_references_used_for_primer_design(chikv_scheme):
    for dir in Direction:
        assert any(
            map(
                lambda x: x.seq != x.reference_msa[0].seq,
                [p for p in chikv_scheme.primers if p.direction == dir],
            )
        )


def test_scheme_runs_with_single_reference(ncov_single_ref_input):
    references = process_fasta(ncov_single_ref_input)
    scheme = MultiplexScheme(references)
    scheme.design_scheme()

    assert len(scheme.regions) > 30


def test_first_reference_is_primary(ebola_input, ebola_scheme):
    records = list(SeqIO.parse(ebola_input, "fasta"))
    assert records[0].id == ebola_scheme.primary_ref.id


def test_first_only_option_has_no_mistmatches_against_primary(chikv_input):
    scheme = get_scheme(chikv_input, primary_only=True)
    for dir in Direction:
        assert all(
            map(
                lambda x: x.seq == x.reference_msa[0].seq,
                [p for p in scheme.primers if p.direction == dir],
            )
        )



import primer3

from Bio.Align import MultipleSeqAlignment
from primalscheme import config
from primalscheme.primer import Primer, Direction


def test_base_penalty_matches_primer3_penalty(random_reference_slice):
    """Check internal penalty calculation matches primer3 output"""
    p3_global = {
        "PRIMER_NUM_RETURN": 10,
        "PRIMER_PRODUCT_SIZE_RANGE": [
            [config.AMPLICON_SIZE_MIN, config.AMPLICON_SIZE_MIN]
        ],
        "PRIMER_OPT_SIZE": config.PRIMER_SIZE_RANGE.opt,
        "PRIMER_MIN_SIZE": config.PRIMER_SIZE_RANGE.min,
        "PRIMER_MAX_SIZE": config.PRIMER_SIZE_RANGE.max,
        "PRIMER_OPT_TM": config.PRIMER_OPT_TM,
        "PRIMER_MIN_TM": config.PRIMER_MIN_TM,
        "PRIMER_MAX_TM": config.PRIMER_MAX_TM,
        "PRIMER_OPT_GC_PERCENT": config.PRIMER_GC_RANGE.opt,
        "PRIMER_MIN_GC": config.PRIMER_GC_RANGE.min,
        "PRIMER_MAX_GC": config.PRIMER_GC_RANGE.max,
        "PRIMER_MAX_POLY_X": config.PRIMER_MAX_HOMO,
        "PRIMER_SALT_MONOVALENT": config.MV_CONC,
        "PRIMER_DNTP_CONC": config.DNTP_CONC,
        "PRIMER_SALT_DIVALENT": config.DV_CONC,
        "PRIMER_DNA_CONC": config.DNA_CONC,
    }

    p3_seq = {
        "SEQUENCE_TEMPLATE": str(random_reference_slice.seq),
        "SEQUENCE_PRIMER_PAIR_OK_REGION_LIST": [-1, -1, -1, -1],
        "SEQUENCE_INCLUDED_REGION": [-1, -1],
    }

    p3_output = primer3.bindings.designPrimers(p3_seq, p3_global)
    text_dir = ("LEFT", "RIGHT")

    for d in range(2):
        num_returned = p3_output[f"PRIMER_{text_dir[d]}_NUM_RETURNED"]
        for i in range(num_returned):
            seq = str(p3_output[f"PRIMER_{text_dir[d]}_{i}_SEQUENCE"])
            p3_penalty = float(p3_output[f"PRIMER_{text_dir[d]}_{i}_PENALTY"])
            mock_msa = MultipleSeqAlignment([random_reference_slice])
            direction = Direction.LEFT if text_dir[d] == "LEFT" else Direction.RIGHT
            primer = Primer(seq, 0, direction, 1, mock_msa)

            assert primer.base_penalty == p3_penalty



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['primalscheme']

package_data = \
{'': ['*']}

package_dir = \
{'': 'src'}

install_requires = \
['biopython>=1,<2',
 'click==7.1.2',
 'parasail==1.2',
 'primer3-py>=0,<1',
 'progress==1.5',
 'reportlab>=3,<4']

extras_require = \
{'test': ['pytest >=2.7.3', 'pytest-cov']}

entry_points = \
{'console_scripts': ['primalscheme = primalscheme.cli:cli']}

setup(name='primalscheme',
      version='1.3.1',
      description='primalscheme is a tool for designing primer panels for multiplex PCR',
      author='Josh Quick',
      author_email='j.quick@bham.ac.uk',
      url='https://github.com/aresti/primalscheme',
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.6,<4',
     )



"""
PrimalScheme: a primer3 wrapper for designing multiplex primer schemes

Copyright (C) 2020 Joshua Quick and Andrew Smith
www.github.com/aresti/primalscheme

This module contains alignment functions.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>
"""

import parasail

from collections import namedtuple

from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord


Alignment = namedtuple("Alignment", "mismatches formatted_alignment")


class FailedAlignmentError(Exception):
    """No aligment between the primer and the refernce"""

    def __init__(self, message, reference=None):
        """Init FailedAlignmentError."""
        self.message = message
        self.reference = reference

    def __str__(self):
        """Exception string representation."""
        return self.message


def align_secondary_reference(primary_flank, secondary_ref):
    """
    An seqan alignment of a primary reference slice against a complete
    secondary reference.
    """

    MATRIX = parasail.matrix_create("ACGT", 2, -1)
    OPEN = 2
    EXTEND = 1

    # Semi-Global, do not penalize gaps at beginning and end of s2/database
    trace = parasail.sg_dx_trace_striped_sat(
        str(primary_flank.seq), str(secondary_ref.seq), OPEN, EXTEND, MATRIX
    )
    traceback = trace.get_traceback()
    query_end = trace.end_query
    ref_end = trace.end_ref
    aligned_query = traceback.query[ref_end - query_end : ref_end + 1]
    aligned_ref = traceback.ref[ref_end - query_end : ref_end + 1]

    # Alignment failed (indels)
    if "-" in aligned_query + aligned_ref or len(primary_flank) != len(aligned_ref):
        raise FailedAlignmentError(
            "Alignment failed between primary and secondary reference.",
            reference=secondary_ref,
        )

    return SeqRecord(Seq(aligned_ref), id=secondary_ref.id)
